#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import inspect
import logging

from connect.eaas.core.models import (
    Context,
)
from connect.eaas.runner.config import (
    ConfigHelper,
)
from connect.eaas.runner.handlers.base import (
    ApplicationHandlerBase,
)


logger = logging.getLogger(__name__)


class TfnApp(ApplicationHandlerBase):
    """
    Handle the lifecycle of a Transformation extension.
    """
    LOGGER_NAME = 'eaas.tfnapp'

    def __init__(self, config: ConfigHelper):
        super().__init__(config)
        self._config = config
        self._transformations = None

    def get_application(self):
        return self.load_application('tfnapp')

    @property
    def transformations(self):
        return self.get_application().get_transformations()

    def get_features(self):
        return {
            'transformations': self.transformations,
        }

    def get_method(
            self,
            event_type,
            task_id,
            method_name,
            transformation_request=None,
            installation=None,
            api_key=None,
            connect_correlation_id=None,
    ):
        if not method_name:  # pragma: no cover
            return

        args = (
            self._create_client(
                event_type, task_id, method_name, self._config.api_key, connect_correlation_id,
            ),
            self.get_logger(extra={'task_id': task_id}),
            self._config.variables,
        )

        kwargs = {}
        if installation:
            kwargs['installation'] = installation
            kwargs['installation_client'] = self._create_client(
                event_type,
                task_id,
                method_name,
                api_key,
                connect_correlation_id,
            )
        if transformation_request:
            kwargs['transformation_request'] = transformation_request

        app_class = self.get_application()

        if 'context' in inspect.signature(app_class.__init__).parameters:
            kwargs['context'] = Context(
                extension_id=self.config.service_id,
                environment_id=self.config.environment_id,
                environment_type=self.config.environment_type,
            )

        ext = app_class(*args, **kwargs)

        return getattr(ext, method_name, None)



import asyncio


class ResultStore:
    def __init__(self):
        self.lock = asyncio.Lock()
        self.futures = {}

    async def put(self, idx, data):
        async with self.lock:
            future = self.futures.setdefault(idx, asyncio.Future())
            future.set_result(data)

    async def get(self, idx):
        async with self.lock:
            future = self.futures.setdefault(idx, asyncio.Future())
        data = await future
        async with self.lock:
            del self.futures[idx]
        return data



import logging
import signal
import sys
import threading
import time
from collections import (
    namedtuple,
)
from multiprocessing import (
    get_context,
)
from pathlib import (
    Path,
)

from rich.console import (
    Console,
)
from watchfiles import (
    watch,
)
from watchfiles.filters import (
    PythonFilter,
)
from watchfiles.run import (
    start_process,
)

from connect.eaas.runner.artworks.banner import (
    print_banner,
)
from connect.eaas.runner.config import (
    ConfigHelper,
)
from connect.eaas.runner.constants import (
    ANVILAPP_WORKER,
    BACKGROUND_EVENTS_WORKER,
    INTERACTIVE_EVENTS_WORKER,
    PROCESS_CHECK_INTERVAL_SECS,
    TFNAPP_WORKER,
    WEBAPP_WORKER,
    WORKER_TYPES,
)
from connect.eaas.runner.handlers.anvil import (
    AnvilApp,
)
from connect.eaas.runner.handlers.events import (
    EventsApp,
)
from connect.eaas.runner.handlers.transformations import (
    TfnApp,
)
from connect.eaas.runner.handlers.web import (
    WebApp,
)
from connect.eaas.runner.helpers import (
    get_features_table,
    get_no_features_table,
    notify_process_restarted,
    validate_extension,
)
from connect.eaas.runner.workers.anvil import (
    start_anvilapp_worker_process,
)
from connect.eaas.runner.workers.events import (
    start_background_worker_process,
    start_interactive_worker_process,
)
from connect.eaas.runner.workers.transformations import (
    start_tfnapp_worker_process,
)
from connect.eaas.runner.workers.web import (
    start_webapp_worker_process,
)


logger = logging.getLogger('connect.eaas')


LifecycleFlags = namedtuple('LifecycleFlags', ('on_startup', 'on_shutdown'))


HANDLED_SIGNALS = (signal.SIGINT, signal.SIGTERM)


def _display_path(path):
    try:
        return f'"{path.relative_to(Path.cwd())}"'
    except ValueError:  # pragma: no cover
        return f'"{path}"'


class Master:

    HANDLER_CLASSES = {
        BACKGROUND_EVENTS_WORKER: EventsApp,
        INTERACTIVE_EVENTS_WORKER: EventsApp,
        WEBAPP_WORKER: WebApp,
        ANVILAPP_WORKER: AnvilApp,
        TFNAPP_WORKER: TfnApp,
    }

    PROCESS_TARGETS = {
        BACKGROUND_EVENTS_WORKER: start_background_worker_process,
        INTERACTIVE_EVENTS_WORKER: start_interactive_worker_process,
        WEBAPP_WORKER: start_webapp_worker_process,
        ANVILAPP_WORKER: start_anvilapp_worker_process,
        TFNAPP_WORKER: start_tfnapp_worker_process,
    }

    def __init__(self, secure=True, debug=False, no_rich=False, no_validate=False, reload=False):
        self.config = ConfigHelper(secure=secure)
        self.handlers = {
            worker_type: self.HANDLER_CLASSES[worker_type](self.config)
            for worker_type in WORKER_TYPES
        }
        self.reload = reload
        self.debug = debug
        self.no_rich = no_rich
        self.no_validate = no_validate
        self.workers = {}
        self.stop_event = threading.Event()
        self.monitor_event = threading.Event()
        self.watch_filter = PythonFilter(ignore_paths=None)
        self.watcher = watch(
            Path.cwd(),
            watch_filter=self.watch_filter,
            stop_event=self.stop_event,
            yield_on_timeout=True,
        )
        self.monitor_thread = None
        self.lifecycle_lock = get_context('spawn').Lock()
        self.lifecycle_events = {
            handler_class: LifecycleFlags(
                get_context('spawn').Value('H', 0),
                get_context('spawn').Value('H', 0),
            )
            for handler_class in set(self.HANDLER_CLASSES.values())
        }
        self.setup_signals_handler()

    def setup_signals_handler(self):
        for sig in HANDLED_SIGNALS:
            signal.signal(sig, self.handle_signal)

    def get_available_features(self):
        have_features = False
        features = {}
        for handler in self.handlers.values():
            have_features = have_features or handler.should_start
            worker_info = {
                'available': handler.should_start,
            }
            if handler.should_start:
                worker_info['features'] = handler.features
            features[handler.__class__.__name__] = worker_info
        return have_features, features

    def handle_signal(self, *args, **kwargs):
        self.stop_event.set()

    def check(self):
        console = Console()

        highest_message_level = 'INFO'
        tables = []

        if not self.no_validate:
            highest_message_level, tables = validate_extension()

        have_features, features = self.get_available_features()

        if not have_features:
            highest_message_level = 'ERROR'

        console.print()

        print_banner(highest_message_level)

        if not have_features:
            console.print(get_no_features_table())

        for table in tables:
            console.print(table)

        if highest_message_level == 'ERROR':
            return False

        console.print()
        console.print(get_features_table(features))
        return True

    def start(self):
        if not self.check():
            sys.exit(-1)
        for worker_type, handler in self.handlers.items():
            if handler.should_start:
                self.start_worker_process(worker_type, handler)
        self.monitor_thread = threading.Thread(target=self.monitor_processes)
        self.monitor_event.set()
        self.monitor_thread.start()

    def start_worker_process(self, worker_type, handler):
        lifecycle_flags = self.lifecycle_events[handler.__class__]
        p = start_process(
            self.PROCESS_TARGETS[worker_type],
            'function',
            (
                handler.__class__,
                self.config,
                self.lifecycle_lock,
                lifecycle_flags.on_startup,
                lifecycle_flags.on_shutdown,
                self.debug,
                self.no_rich,
            ),
            {},
        )
        self.workers[worker_type] = p
        logger.info(f'{worker_type.capitalize()} worker pid: {p.pid}')

    def monitor_processes(self):
        while self.monitor_event.is_set():
            exited_workers = []
            for worker_type, p in self.workers.items():
                if not p.is_alive():
                    if p.exitcode != 0:
                        notify_process_restarted(worker_type)
                        logger.info(f'Process of type {worker_type} is dead, restart it')
                        self.start_worker_process(worker_type, self.handlers[worker_type])
                    else:
                        exited_workers.append(worker_type)
                        logger.info(f'{worker_type.capitalize()} worker exited')
            if exited_workers == list(self.workers.keys()):
                self.stop_event.set()

            time.sleep(PROCESS_CHECK_INTERVAL_SECS)

    def stop(self):
        self.monitor_event.clear()
        self.monitor_thread.join()
        for process in self.workers.values():
            process.stop(sigint_timeout=5, sigkill_timeout=1)
            logger.info(f'Consumer process with pid {process.pid} stopped.')

    def restart(self):
        self.stop()
        for handler in self.handlers.values():
            handler.reload()
        self.start()

    def __iter__(self):
        return self

    def __next__(self):
        changes = next(self.watcher)
        if changes:
            return list({Path(c[1]) for c in changes})
        return None

    def run(self):
        self.start()
        if self.reload:
            for files_changed in self:
                if files_changed:
                    logger.warning(
                        'Detected changes in %s. Reloading...',
                        ', '.join(map(_display_path, files_changed)),
                    )
                    self.restart()
        else:
            self.stop_event.wait()
        self.stop()



#
# Copyright (c) 2023 Ingram Micro. All Rights Reserved.
#

import functools
import inspect
import logging

from fastapi import (
    FastAPI,
)
from fastapi.middleware.cors import (
    CORSMiddleware,
)
from fastapi.openapi.utils import (
    generate_operation_summary,
    get_openapi,
)
from fastapi.staticfiles import (
    StaticFiles,
)
from starlette.middleware.base import (
    BaseHTTPMiddleware,
)

from connect.client import (
    ClientError,
)
from connect.eaas.core.decorators import (
    router as root_router,
)
from connect.eaas.core.utils import (
    client_error_exception_handler,
)
from connect.eaas.runner.config import (
    ConfigHelper,
)
from connect.eaas.runner.handlers.base import (
    ApplicationHandlerBase,
)


logger = logging.getLogger(__name__)


class _OpenApiCORSMiddleware(CORSMiddleware):
    async def __call__(self, scope, receive, send):
        if scope['type'] != 'http' or scope['path'] != '/openapi/spec.json':  # pragma: no cover
            await self.app(scope, receive, send)
            return

        await super().__call__(scope, receive, send)  # pragma: no cover


class _ProxyHeadersMiddleware:
    def __init__(self, app):
        self.app = app

    async def __call__(self, scope, receive, send):
        if scope['type'] == 'http':  # pragma: no branch
            headers = dict(scope['headers'])
            if b'x-forwarded-host' in headers:
                headers[b'host'] = headers[b'x-forwarded-host']
                scope['headers'] = list(headers.items())
            if b'x-forwarded-proto' in headers:
                scope['scheme'] = headers[b'x-forwarded-proto'].decode('ascii')

        await self.app(scope, receive, send)


class WebApp(ApplicationHandlerBase):
    """
    Handle the lifecycle of an extension.
    """

    LOGGER_NAME = 'eaas.webapp'

    def __init__(self, config: ConfigHelper):
        super().__init__(config)
        self._logging_handler = None
        self._app = None

    @property
    def ui_modules(self):
        return self.get_application().get_ui_modules()

    @property
    def proxied_connect_api(self):
        return self.get_application().get_proxied_connect_api()

    @property
    def app(self):
        if not self._app:
            self._app = self.get_asgi_application()
        return self._app

    def get_application(self):
        return self.load_application('webapp')

    def get_features(self):
        return {
            'endpoints': self.get_endpoints(),
            'ui_modules': self.get_ui_components(),
        }

    def get_ui_components(self):
        components = []

        if 'settings' in self.ui_modules:
            components.append({
                'name': self.ui_modules['settings']['label'],
                'url': self.ui_modules['settings']['url'],
                'integration_point': 'Account Settings',
            })

        if 'modules' in self.ui_modules:
            components.append({
                'name': self.ui_modules['modules']['label'],
                'url': self.ui_modules['modules']['url'],
                'integration_point': 'Module Feature',
            })
            for child in self.ui_modules['modules'].get('children', []):
                components.append({
                    'name': child['label'],
                    'url': child['url'],
                    'integration_point': 'Module Feature',
                })

        for admin in self.ui_modules.get('admins', []):
            components.append({
                'name': admin['label'],
                'url': admin['url'],
                'integration_point': 'Installations Admin',
            })

        for customer in self.ui_modules.get('customer', []):
            components.append({
                'name': customer['label'],
                'url': customer['url'],
                'icon': customer['icon'],
                'integration_point': 'Customer Home Page',
            })

        return components

    def get_endpoints(self):
        auth, no_auth = self.get_application().get_routers()
        endpoints = {
            'auth': [],
            'no_auth': [],
        }
        for router, group in ((auth, 'auth'), (no_auth, 'no_auth')):
            for route in router.routes:
                for method in route.methods:
                    endpoints[group].append({
                        'summary': generate_operation_summary(route=route, method=method),
                        'method': method,
                        'path': route.path,
                    })
        return endpoints

    def get_asgi_application(self):
        exception_handlers = {ClientError: client_error_exception_handler}
        if hasattr(self.get_application(), 'get_exception_handlers'):
            exception_handlers = self.get_application().get_exception_handlers(
                exception_handlers,
            )

        app = FastAPI(
            openapi_url='/openapi/spec.json',
            exception_handlers=exception_handlers,
        )
        app.add_middleware(
            _OpenApiCORSMiddleware,
            allow_origins=['*'],
        )
        app.add_middleware(_ProxyHeadersMiddleware)

        if hasattr(self.get_application(), 'get_middlewares'):
            self.setup_middlewares(app, self.get_application().get_middlewares() or [])

        auth, no_auth = self.get_application().get_routers()
        app.include_router(auth, prefix='/api')
        app.include_router(no_auth, prefix='/guest')
        app.include_router(no_auth, prefix='/unauthorized')
        app.openapi = functools.partial(self.get_api_schema, app)
        static_root = self.get_application().get_static_root()
        if static_root:
            app.mount('/static', StaticFiles(directory=static_root), name='static')

        return app

    def setup_middlewares(self, app, middlewares):
        for middleware in middlewares:
            if inspect.isfunction(middleware):
                app.add_middleware(BaseHTTPMiddleware, dispatch=middleware)
            elif inspect.isclass(middleware):
                app.add_middleware(middleware)
            elif isinstance(middleware, tuple):
                app.add_middleware(middleware[0], **middleware[1])

    def get_api_schema(self, app):
        if app.openapi_schema:
            return app.openapi_schema

        openapi_schema = get_openapi(
            title=self.name,
            description=self.description,
            version=self.version,
            routes=app.routes,
        )

        paths = openapi_schema.get('paths')

        if paths:
            for path in paths.values():
                for operation in path.values():
                    params = []
                    for param in operation.get('parameters', []):
                        if param['in'] == 'header' and param['name'].startswith('x-connect-'):
                            continue
                        params.append(param)
                    operation['parameters'] = params

        app.openapi_schema = openapi_schema
        return app.openapi_schema

    def reload(self):
        root_router.routes = []
        return super().reload()



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import asyncio
import logging
import signal

from devtools import (
    pformat,
)

from connect.eaas.core.proto import (
    Message,
    MessageType,
    SetupRequest,
)
from connect.eaas.runner.constants import (
    DELAY_ON_CONNECT_EXCEPTION_SECONDS,
    RESULT_SENDER_MAX_RETRIES,
    RESULT_SENDER_WAIT_GRACE_SECONDS,
)
from connect.eaas.runner.helpers import (
    configure_logger,
    get_version,
)
from connect.eaas.runner.managers.transformation import (
    TransformationTasksManager,
)
from connect.eaas.runner.workers.base import (
    WorkerBase,
)


logger = logging.getLogger(__name__)


class TransformationWorker(WorkerBase):
    """
    The TransformationWorker is responsible to handle the websocket connection
    with the server. It will send found transformations to
    the server and wait for tasks that need to be processed using
    the task manager.
    """
    def __init__(self, handler, lock, startup_fired, shutdown_fired, runner_type=None):
        super().__init__(handler, lock, startup_fired, shutdown_fired)
        self.runner_type = runner_type
        self.results_queue = asyncio.Queue()
        self.manager = TransformationTasksManager(
            self.config,
            self.handler,
            self.results_queue.put,
        )

    def get_url(self):
        return self.config.get_tfnapp_ws_url()

    def get_setup_request(self):
        msg = Message(
            version=2,
            message_type=MessageType.SETUP_REQUEST,
            data=SetupRequest(
                repository={
                    'readme_url': self.handler.readme,
                    'changelog_url': self.handler.changelog,
                },
                audience=self.handler.audience,
                transformations=self.handler.transformations,
                runner_version=get_version(),
            ),
        )
        logger.debug(f'Sending setup request: {pformat(msg)}')
        return msg.dict()

    async def stopping(self):
        timeout = self.config.get_timeout('transformation') + RESULT_SENDER_WAIT_GRACE_SECONDS
        try:
            await asyncio.wait_for(
                self.results_task,
                timeout=timeout,
            )
        except asyncio.TimeoutError:
            logger.error(
                f'Cannot send all results timeout of {timeout} exceeded, cancel task',
            )
            self.results_task.cancel()
            try:
                await self.results_task
            except asyncio.CancelledError:
                logger.info('Result sender task has been cancelled')

    async def process_message(self, data):
        message = Message.deserialize(data)
        logger.debug(f'Received message: {pformat(message)}')
        if message.message_type == MessageType.SETUP_RESPONSE:
            await self.process_setup_response(message.data)
        elif message.message_type == MessageType.TASK:
            logger.info(
                f'Received new transformation {message.data.options.task_id}',
            )
            await self.manager.submit(message.data)
            logger.info(f'Task {message.data.options.task_id} submitted for processing')
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def result_sender(self):  # noqa: CCR001
        """
        Dequeues results from the results queue and send it to
        the EaaS backend.
        """
        await self.run_event.wait()
        while True:
            if self.results_queue.empty():
                if not self.run_event.is_set() and self.manager.running_tasks == 0:
                    logger.info(f'{self} exiting and no more running tasks: exit!')
                    return
                await asyncio.sleep(.5)
                continue
            logger.info(
                f'Current transformation processing status: '
                f'running={self.manager.running_tasks}, results={self.results_queue.qsize()}',
            )
            result = await self.results_queue.get()
            logger.info(f'Got a transformation result from queue: {result.options.task_id}')
            retries = 0
            while retries < RESULT_SENDER_MAX_RETRIES:
                try:
                    message = Message(
                        version=2,
                        message_type=MessageType.TASK,
                        data=result,
                    )
                    await self.ensure_connection()
                    logger.debug(f'Sending message: {pformat(message)}')
                    await self.send(message.serialize())
                    logger.info(
                        f'Result for transformation task {result.options.task_id} has been sent.',
                    )
                    break
                except Exception:
                    logger.warning(
                        f'Attempt {retries} to send results for task '
                        f'{result.options.task_id} has failed.',
                    )
                    retries += 1
                    await asyncio.sleep(DELAY_ON_CONNECT_EXCEPTION_SECONDS)
            else:
                logger.warning(
                    f'Max retries exceeded ({RESULT_SENDER_MAX_RETRIES})'
                    f' for sending results of transformation task {result.options.task_id}',
                )

            if not self.run_event.is_set():
                logger.info(
                    f'Current transformation processing status: '
                    f'running={self.manager.running_tasks}, results={self.results_queue.qsize()}',
                )

    async def start(self):
        """
        Start the runner.
        """
        self.results_task = asyncio.create_task(self.result_sender())
        await super().start()


def start_tfnapp_worker_process(
    handler_class,
    config,
    lifecycle_lock,
    on_startup_fired,
    on_shutdown_fired,
    debug,
    no_rich,
):
    handler = handler_class(config)
    configure_logger(debug, no_rich)
    worker = TransformationWorker(
        handler,
        lifecycle_lock,
        on_startup_fired,
        on_shutdown_fired,
    )
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(
        signal.SIGINT,
        worker.handle_signal,
    )
    loop.add_signal_handler(
        signal.SIGTERM,
        worker.handle_signal,
    )
    loop.run_until_complete(worker.start())



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import logging
import os
import subprocess
from datetime import (
    datetime,
)
from importlib.metadata import (
    entry_points,
    version,
)
from uuid import (
    uuid4,
)

import requests
from rich import (
    box,
)
from rich.align import (
    Align,
)
from rich.markdown import (
    Markdown,
)
from rich.syntax import (
    Syntax,
)
from rich.table import (
    Table,
)

from connect.client import (
    ClientError,
    ConnectClient,
)
from connect.eaas.core.validation.models import (
    ValidationItem,
    ValidationResult,
)
from connect.eaas.core.validation.validators import (
    get_validators,
)
from connect.eaas.runner.constants import (
    BACKGROUND_TASK_MAX_EXECUTION_TIME,
    HANDLER_CLASS_TITLE,
    INTERACTIVE_TASK_MAX_EXECUTION_TIME,
    ORDINAL_SUFFIX,
    PYPI_EXTENSION_RUNNER_URL,
    ROW_TRANSFORMATION_TASK_MAX_EXECUTION_TIME,
    SCHEDULED_TASK_MAX_EXECUTION_TIME,
    TRANSFORMATION_TASK_MAX_EXECUTION_TIME,
    TRANSFORMATION_WRITE_QUEUE_TIMEOUT,
)


logger = logging.getLogger('connect.eaas')


def get_container_id():
    result = subprocess.run(
        ['cat', '/proc/1/cpuset'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
    except subprocess.CalledProcessError:
        return str(uuid4())

    _, container_id = result.stdout.decode()[:-1].rsplit('/', 1)
    if len(container_id) == 64:
        return container_id

    result = subprocess.run(
        ['grep', 'overlay', '/proc/self/mountinfo'],
        capture_output=True,
        stdin=subprocess.DEVNULL,
        start_new_session=True,
    )
    try:
        result.check_returncode()
        mount = result.stdout.decode()
        start_idx = mount.index('upperdir=') + len('upperdir=')
        end_idx = mount.index(',', start_idx)
        dir_path = mount[start_idx:end_idx]
        _, container_id, _ = dir_path.rsplit('/', 2)
        if len(container_id) != 64:
            return str(uuid4())
        return container_id
    except (subprocess.CalledProcessError, ValueError):
        return str(uuid4())


def get_environment():
    return {
        'api_key': os.getenv('API_KEY'),
        'environment_id': os.getenv('ENVIRONMENT_ID'),
        'instance_id': os.getenv('INSTANCE_ID', get_container_id()),
        'ws_address': os.getenv('SERVER_ADDRESS', 'api.cnct.info'),
        'api_address': os.getenv('API_ADDRESS', os.getenv('SERVER_ADDRESS', 'api.cnct.info')),
        'background_task_max_execution_time': int(os.getenv(
            'BACKGROUND_TASK_MAX_EXECUTION_TIME', BACKGROUND_TASK_MAX_EXECUTION_TIME,
        )),
        'interactive_task_max_execution_time': int(os.getenv(
            'INTERACTIVE_TASK_MAX_EXECUTION_TIME', INTERACTIVE_TASK_MAX_EXECUTION_TIME,
        )),
        'scheduled_task_max_execution_time': int(os.getenv(
            'SCHEDULED_TASK_MAX_EXECUTION_TIME', SCHEDULED_TASK_MAX_EXECUTION_TIME,
        )),
        'transformation_task_max_execution_time': int(os.getenv(
            'TRANSFORMATION_TASK_MAX_EXECUTION_TIME', TRANSFORMATION_TASK_MAX_EXECUTION_TIME,
        )),
        'row_transformation_task_max_execution_time': int(
            os.getenv(
                'ROW_TRANSFORMATION_TASK_MAX_EXECUTION_TIME',
                ROW_TRANSFORMATION_TASK_MAX_EXECUTION_TIME,
            ),
        ),
        'transformation_write_queue_timeout': int(os.getenv(
            'TRANSFORMATION_WRITE_QUEUE_TIMEOUT', TRANSFORMATION_WRITE_QUEUE_TIMEOUT,
        )),
    }


def get_version():
    try:
        return version('connect-extension-runner')
    except Exception:
        return '0.0.0'


def to_ordinal(val):
    if val > 14:
        return f"{val}{ORDINAL_SUFFIX.get(int(str(val)[-1]), 'th')}"
    return f"{val}{ORDINAL_SUFFIX.get(val, 'th')}"


def get_connect_version():
    connect_client = get_client()
    try:
        connect_client.accounts.all().first()
    except ClientError as ce:
        if ce.status_code in [401, 403]:
            raise Exception(
                'Cannot check the current EaaS Runner version: '
                f'API key is not valid: {ce}.')
        raise Exception(f'Cannot check the current EaaS Runner version: {ce}.')

    return connect_client.response.headers['Connect-Version']


def get_pypi_runner_latest_version(major_version):
    res = requests.get(PYPI_EXTENSION_RUNNER_URL)
    if res.status_code != 200:
        logger.error(
            f'Cannot check the current EaaS Runner version: {res.text}.',
        )
        raise Exception(f'Cannot check the current EaaS Runner version: {res.text}.')

    content = res.json()
    tags = [
        int(version.split('.')[1])
        for version in content['releases'] if version.startswith(f'{major_version}.')
    ]
    if tags:
        return f'{major_version}.{str(max(tags))}'
    return content['info']['version']


def get_client():
    env = get_environment()
    api_key = env['api_key']
    api_address = env['api_address']
    return ConnectClient(
        api_key,
        endpoint=f'https://{api_address}/public/v1',
        use_specs=False,
        max_retries=3,
    )


def get_current_environment():
    env = get_environment()
    environment_id = env['environment_id']
    if environment_id.startswith('ENV-'):
        service_id = f'SRVC-{environment_id[4:-3]}'
        client = get_client()
        try:
            return (
                client('devops').services[service_id].environments[environment_id].get()
            )
        except ClientError as ce:
            logger.warning(f'Cannot retrieve environment information: {ce}')


def notify_process_restarted(process_type):
    env = get_environment()
    current_environment = get_current_environment()
    if current_environment and current_environment['runtime'] == 'cloud':
        instance_id = env['instance_id']
        environment_id = current_environment['id']
        service_id = f'SRVC-{environment_id[4:-3]}'
        client = get_client()
        try:
            client('devops').services[service_id].environments[environment_id].update(
                {
                    'error_output': (
                        f'Process {process_type} worker of instance '
                        f'{instance_id} has been '
                        f'restarted at {datetime.now().isoformat()}'
                    ),
                },
            )
        except ClientError as ce:
            logger.warning(f'Cannot notify {process_type} process restart: {ce}')


def iter_entry_points(group, name=None):
    group_entrypoints = entry_points().get(group)
    if not group_entrypoints:
        return
    for ep in group_entrypoints:
        if name:
            if ep.name == name:
                yield ep
        else:
            yield ep


def configure_logger(debug, no_rich):
    logging.config.dictConfig(
        {
            'version': 1,
            'disable_existing_loggers': False,
            'formatters': {
                'verbose': {
                    'format': '%(asctime)s %(name)s %(levelname)s PID_%(process)d %(message)s',
                },
                'rich': {
                    'format': '%(message)s',
                },
            },
            'filters': {},
            'handlers': {
                'console': {
                    'class': 'logging.StreamHandler',
                    'formatter': 'verbose',
                },
                'rich': {
                    'class': 'rich.logging.RichHandler',
                    'formatter': 'rich',
                    'log_time_format': lambda x: x.strftime('%Y-%m-%d %H:%M:%S.%f')[:-3],
                    'rich_tracebacks': True,

                },
            },
            'loggers': {
                'connect.eaas': {
                    'handlers': ['console'] if no_rich else ['rich'],
                    'level': 'DEBUG' if debug else 'INFO',
                },
                'eaas': {
                    'handlers': ['console'] if no_rich else ['rich'],
                    'level': 'DEBUG' if debug else 'INFO',
                },
            },
        },
    )


def check_runner_version(context):
    try:
        connect_full_version = get_connect_version()
        connect_version = connect_full_version.split('.')[0]
        runner_version = get_version()
        latest_runner_version = get_pypi_runner_latest_version(connect_version)

        if runner_version != latest_runner_version:
            return ValidationResult(
                items=[
                    ValidationItem(
                        level='ERROR',
                        message=(
                            'Runner is outdated, please, update. '
                            f'Required version {latest_runner_version}, '
                            f'current version: {runner_version}.'
                        ),
                    ),
                ],
                must_exit=True,
            )
    except Exception as e:
        return ValidationResult(
            items=[
                ValidationItem(level='ERROR', message=str(e)),
            ],
            must_exit=True,
        )
    return ValidationResult()


def validate_extension():
    validation_items = []
    event_definitions = {
        definition['type']: definition
        for definition in get_client()('devops').event_definitions.all()
    }
    context = {
        'runner_version': get_version(),
        'project_dir': os.getcwd(),
        'event_definitions': event_definitions,
    }
    for validator in [check_runner_version, *get_validators()]:
        result = validator(context=context)
        validation_items.extend(result.items)
        if result.must_exit:
            break
        if result.context:
            context.update(result.context)

    higest_message_level = 'INFO'

    tables = []

    if validation_items:
        higest_message_level = 'WARNING'
        for item in validation_items:
            if item.level == 'ERROR':
                higest_message_level = 'ERROR'
            table = Table(
                box=box.ROUNDED,
                show_header=False,
            )
            table.add_column('Field', style='blue')
            table.add_column('Value', overflow='fold')
            level_color = 'red' if item.level == 'ERROR' else 'yellow'
            table.add_row('Level', f'[bold {level_color}]{item.level}[/]')
            table.add_row('Message', Markdown(item.message))
            table.add_row('File', item.file or '-')
            table.add_row(
                'Code',
                Syntax(
                    item.code,
                    'python3',
                    theme='ansi_light',
                    dedent=True,
                    line_numbers=True,
                    start_line=item.start_line,
                    highlight_lines={item.lineno},
                ) if item.code else '-',
            )
            tables.append(table)
    return higest_message_level, tables


def get_no_features_table():
    table = Table(
        box=box.ROUNDED,
        show_header=False,
    )
    table.add_column('Field', style='blue')
    table.add_column('Value')
    table.add_row('Level', '[bold red]ERROR[/]')
    table.add_row(
        'Message',
        (
            'No feature available for current extension. '
            'Did you run poetry install ?'
        ),
    )
    table.add_row('File', '-')
    table.add_row('Code', '-')
    return table


def get_anvilapp_detail_table(details):
    callables_table = None

    callables = details['features'].get('callables')

    if callables:
        callables_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Anvil Callables',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        callables_table.add_column('Method')
        callables_table.add_column('Summary')
        callables_table.add_column('Description')
        for callable in callables:
            callables_table.add_row(
                callable['method'],
                callable['summary'],
                callable['description'],
            )

    return callables_table


def get_eventsapp_detail_table(details):
    events_table = None
    schedulables_table = None
    grid = None

    events = details['features'].get('events')
    schedulables = details['features'].get('schedulables')

    if events:
        events_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Event Subscriptions',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        events_table.add_column('Method Name')
        events_table.add_column('Event Type')
        events_table.add_column('Statuses')
        for event in events.values():
            events_table.add_row(
                event['method'],
                event['event_type'],
                ', '.join(event['statuses'] or []),
            )

    if schedulables:
        schedulables_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Schedulable Methods',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        schedulables_table.add_column('Method Name')
        schedulables_table.add_column('Name')
        schedulables_table.add_column('Description')
        for schedulable in schedulables:
            schedulables_table.add_row(
                schedulable['method'],
                schedulable['name'],
                schedulable['description'],
            )

    if events_table and schedulables_table:
        grid = Table.grid(expand=True)
        grid.add_column()
        grid.add_row(events_table)
        grid.add_row(schedulables_table)

    return grid or events_table or schedulables_table


def get_webapp_detail_table(details):
    auth_table = None
    no_auth_table = None
    ui_modules_table = None
    grid = None

    auth_endpoints = details['features']['endpoints'].get('auth')
    no_auth_endpoints = details['features']['endpoints'].get('no_auth')
    ui_modules = details['features'].get('ui_modules')

    if auth_endpoints:
        auth_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Authenticated API Endpoints',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        auth_table.add_column('Summary')
        auth_table.add_column('Method')
        auth_table.add_column('Path')
        for endpoint in auth_endpoints:
            auth_table.add_row(
                endpoint['summary'],
                endpoint['method'],
                endpoint['path'],
            )

    if no_auth_endpoints:
        no_auth_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Non Authenticated API Endpoints',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        no_auth_table.add_column('Summary')
        no_auth_table.add_column('Method')
        no_auth_table.add_column('Path')
        for endpoint in no_auth_endpoints:
            no_auth_table.add_row(
                endpoint['summary'],
                endpoint['method'],
                endpoint['path'],
            )

    if ui_modules:
        ui_modules_table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='UI Components',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        ui_modules_table.add_column('Name')
        ui_modules_table.add_column('Url')
        ui_modules_table.add_column('Integration point')
        for ui_module in ui_modules:
            ui_modules_table.add_row(
                ui_module['name'],
                ui_module['url'],
                ui_module['integration_point'],
            )

    grid = Table.grid(expand=True)
    grid.add_column()
    if auth_table:
        grid.add_row(auth_table)
    if no_auth_table:
        grid.add_row(no_auth_table)
    if ui_modules_table:
        grid.add_row(ui_modules_table)

    return grid


def get_tfnapp_detail_table(details):
    grid = Table.grid(expand=True)
    grid.add_column()

    transformations = details['features'].get('transformations')

    if transformations:
        table = Table(
            box=box.MINIMAL_HEAVY_HEAD,
            title='Transformations',
            title_style='blue',
            show_header=True,
            expand=True,
            row_styles=['', 'dim'],
        )
        table.add_column('Name')
        table.add_column('Description')
        table.add_column('Edit dialog')
        table.add_column('Method')
        for transformation in transformations:
            table.add_row(
                transformation['name'],
                transformation['description'],
                transformation['edit_dialog_ui'],
                transformation['method'],
            )

        grid.add_row(table)

    return grid


def get_features_table(features):
    table = Table(
        box=box.ROUNDED,
        title='Available Features',
        title_style='blue',
        show_header=True,
        show_lines=True,
        expand=True,
    )
    table.add_column('Feature Name')
    table.add_column('Supported', width=9)
    table.add_column('Details')

    for feature_handler, details in features.items():
        if not details['available']:
            table.add_row(
                HANDLER_CLASS_TITLE[feature_handler],
                Align.center('[red]\u2716[/]'),
                '-',
            )
            continue
        if feature_handler == 'EventsApp':
            table.add_row(
                HANDLER_CLASS_TITLE[feature_handler],
                Align.center('[green]\u2714[/]'),
                Align.center(get_eventsapp_detail_table(details)),
            )
        elif feature_handler == 'WebApp':
            table.add_row(
                HANDLER_CLASS_TITLE[feature_handler],
                Align.center('[green]\u2714[/]'),
                Align.center(get_webapp_detail_table(details)),
            )
        elif feature_handler == 'AnvilApp':
            table.add_row(
                HANDLER_CLASS_TITLE[feature_handler],
                Align.center('[green]\u2714[/]'),
                Align.center(get_anvilapp_detail_table(details)),
            )
        elif feature_handler == 'TfnApp':
            table.add_row(
                HANDLER_CLASS_TITLE[feature_handler],
                Align.center('[green]\u2714[/]'),
                Align.center(get_tfnapp_detail_table(details)),
            )
    return table



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import asyncio
import logging
import time
import traceback

from connect.client import (
    AsyncConnectClient,
)
from connect.client.models import (
    AsyncResource,
)
from connect.eaas.core.enums import (
    ResultType,
)
from connect.eaas.core.proto import (
    Task,
    TaskOutput,
)
from connect.eaas.core.responses import (
    BackgroundResponse,
)
from connect.eaas.runner.managers.base import (
    TasksManagerBase,
)


logger = logging.getLogger(__name__)


class BackgroundTasksManager(TasksManagerBase):

    def get_method_name(self, task_data, argument):
        return self.handler.events[task_data.input.event_type]['method']

    def send_skip_response(self, data, output):
        future = asyncio.Future()
        future.set_result(BackgroundResponse.skip(output))
        asyncio.create_task(self.enqueue_result(data, future))

    async def get_argument(self, task_data):
        """
        Get the request object through the Connect public API
        related to the task that need processing.
        """
        client = self.client
        if task_data.options.api_key:
            client = AsyncConnectClient(
                task_data.options.api_key,
                endpoint=self.config.get_api_url(),
                use_specs=False,
                default_headers=self.config.get_user_agent(),
            )

        object_exists = await self.filter_collection_by_event_definition(
            client,
            task_data,
        )
        if not object_exists:
            return

        definition = self.config.event_definitions[task_data.input.event_type]
        url = definition.api_resource_endpoint.format(pk=task_data.input.object_id)
        resource = AsyncResource(client, url)

        return await resource.get()

    async def build_response(self, task_data, future):
        """
        Wait for a background task to be completed and then build the task result message.
        """
        result_message = Task(**task_data.dict())
        result = None
        try:
            begin_ts = time.monotonic()
            result = await asyncio.wait_for(
                future,
                timeout=self.config.get_timeout('background'),
            )
            result_message.output = TaskOutput(result=result.status)
            result_message.output.runtime = time.monotonic() - begin_ts
            logger.info(
                f'background task {task_data.options.task_id} result: {result.status}, took:'
                f' {result_message.output.runtime}',
            )
            if result.status in (ResultType.SKIP, ResultType.FAIL):
                result_message.output.message = result.output

            if result.status == ResultType.RESCHEDULE:
                result_message.output.countdown = result.countdown
        except Exception as e:
            self.log_exception(task_data, e)
            result_message.output = TaskOutput(result=ResultType.RETRY)
            result_message.output.message = traceback.format_exc()[:4000]

        return result_message



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import asyncio
import logging
import signal

from devtools import (
    pformat,
)

from connect.eaas.core.proto import (
    Message,
    MessageType,
    SetupRequest,
)
from connect.eaas.runner.helpers import (
    configure_logger,
    get_version,
)
from connect.eaas.runner.workers.base import (
    WorkerBase,
)


logger = logging.getLogger(__name__)


class AnvilWorker(WorkerBase):
    """
    The EventsWorker is responsible to handle the websocket connection
    with the server. It will send the extension capabilities to
    the server and wait for tasks that need to be processed using
    the tasks manager.
    """

    def get_url(self):
        return self.config.get_anvilapp_ws_url()

    def get_setup_request(self):
        msg = Message(
            version=2,
            message_type=MessageType.SETUP_REQUEST,
            data=SetupRequest(
                variables=self.handler.variables,
                anvil_callables=self.handler.callables,
                repository={
                    'readme_url': self.handler.readme,
                    'changelog_url': self.handler.changelog,
                },
                audience=self.handler.audience,
                runner_version=get_version(),
            ),
        )
        logger.debug(f'Sending setup request: {pformat(msg)}')
        return msg.dict()

    async def stopping(self):
        pass

    async def process_message(self, data):
        message = Message.deserialize(data)
        logger.debug(f'Received message: {pformat(message)}')
        if message.message_type == MessageType.SETUP_RESPONSE:
            await self.process_setup_response(message.data)
            self.handler.start()
        elif message.message_type == MessageType.SHUTDOWN:
            await self.shutdown()

    async def shutdown(self):
        self.handler.stop()
        await super().shutdown()


def start_anvilapp_worker_process(
        handler_class, config, lock, startup_fired, shutdown_fired, debug, no_rich,
):
    handler = handler_class(config)
    configure_logger(debug, no_rich)
    worker = AnvilWorker(handler, lock, startup_fired, shutdown_fired)
    loop = asyncio.get_event_loop()
    loop.add_signal_handler(
        signal.SIGINT,
        worker.handle_signal,
    )
    loop.add_signal_handler(
        signal.SIGTERM,
        worker.handle_signal,
    )
    loop.run_until_complete(worker.start())



#
# This file is part of the Ingram Micro CloudBlue Connect EaaS Extension Runner.
#
# Copyright (c) 2022 Ingram Micro. All Rights Reserved.
#
import asyncio
import inspect
import logging
from abc import (
    ABC,
    abstractmethod,
)
from asyncio.futures import (
    Future,
)
from concurrent.futures import (
    ThreadPoolExecutor,
)
from string import (
    Template,
)
from typing import (
    Union,
)

from connect.client import (
    AsyncConnectClient,
    ClientError,
)
from connect.client.models import (
    AsyncCollection,
)
from connect.eaas.core.enums import (
    TaskCategory,
)
from connect.eaas.runner.config import (
    ConfigHelper,
)
from connect.eaas.runner.handlers.events import (
    EventsApp,
)
from connect.eaas.runner.handlers.transformations import (
    TfnApp,
)


logger = logging.getLogger(__name__)


class TasksManagerBase(ABC):
    """
    Base class for Tasks managers.
    """
    def __init__(self, config: ConfigHelper, handler: Union[EventsApp, TfnApp], enqueue):
        self.config = config
        self.handler = handler
        self.enqueue = enqueue
        self.lock = asyncio.Lock()
        self.executor = ThreadPoolExecutor(max_workers=100)
        self.client = AsyncConnectClient(
            self.config.api_key,
            endpoint=self.config.get_api_url(),
            use_specs=False,
            default_headers=self.config.get_user_agent(),
        )
        self.running_tasks = 0

    async def filter_collection_by_event_definition(self, client, task_data):
        definition = self.config.event_definitions[task_data.input.event_type]
        supported_statuses = []
        template_args = {'_object_id_': task_data.input.object_id}
        if 'events' in self.handler.get_features():
            supported_statuses = self.handler.events[task_data.input.event_type]['statuses']
            template_args['_statuses_'] = f'({",".join(supported_statuses)})'

        rql_filter = Template(definition.api_collection_filter).substitute(
            template_args,
        )

        collection = AsyncCollection(client, definition.api_collection_endpoint)

        count = await collection.filter(rql_filter).count()
        if count == 0:
            logger.info(
                f'Send skip response for {task_data.options.task_id} since '
                'the current request status is not supported.',
            )
            self.send_skip_response(
                task_data,
                'The request status does not match the '
                f'supported statuses: {",".join(supported_statuses)}.',
            )

        return count

    async def submit(self, task_data):
        """
        Submit a new task for its processing.

        :param task_data: Task data.
        :type task_data: connect.eaas.core.proto.Task
        """
        try:
            method = None
            argument = None
            async with self.lock:
                self.running_tasks += 1
            logger.info(
                f'new {task_data.options.task_category} task received: '
                f'{task_data.options.task_id}, running tasks: {self.running_tasks}',
            )
            argument = await self.get_argument(task_data)
            if not argument:
                return
            installation = None
            if task_data.options.installation_id:
                installation = await self.get_installation(task_data)
            method_name = self.get_method_name(task_data, argument)
            kwargs = {
                'installation': installation,
                'api_key': task_data.options.api_key,
                'connect_correlation_id': task_data.options.connect_correlation_id,
            }
            if task_data.options.task_category == TaskCategory.TRANSFORMATION:
                kwargs['transformation_request'] = argument
            method = self.handler.get_method(
                task_data.input.event_type,
                task_data.options.task_id,
                method_name,
                **kwargs,
            )
            if not method:
                async with self.lock:
                    self.running_tasks -= 1
                return

            logger.info(f'invoke method {method.__name__} for task {task_data.options.task_id}')
            await self.invoke(task_data, method, argument)
        except ClientError as ce:
            logger.warning(
                f'Cannot retrieve object {task_data.input.object_id} '
                f'for task {task_data.options.task_id}',
            )
            self.send_exception_response(task_data, ce)
            return
        except Exception as e:
            logger.error(
                f'Exceptions while processing task {task_data.options.task_id}',
            )
            self.send_exception_response(task_data, e)
            return

    def send_exception_response(self, data, e):
        """
        Enqueue a future to send back to EaaS a failed task
        with exception info.
        :param data: Task data.
        :type data: connect.eaas.core.proto.Task
        :param e: The exception that need to be sent back.
        :type e: Exception
        """
        future = Future()
        future.set_exception(e)
        asyncio.create_task(self.enqueue_result(data, future))

    async def enqueue_result(self, task_data, future):
        """
        Build a result object for a task and put it in the
        result queue.
        """
        await self.enqueue(
            await self.build_response(task_data, future),
        )
        async with self.lock:
            self.running_tasks -= 1

    async def invoke(self, task_data, method, argument):
        """
        Invokes a method of the extension class instance
        to process a task. If the method is synchronous
        it will be executed on a ThreadPoolExecutor.

        :param task_data: Data of the task to be processed.
        :type task_data: connect.eaas.core.proto.Task
        :param method: The method that has to be invoked.
        :param argument: The Connect object that need to be processed.
        :type argument: dict
        """
        if inspect.iscoroutinefunction(method):
            future = asyncio.create_task(method(argument))
        else:
            future = asyncio.get_running_loop().run_in_executor(
                self.executor,
                method,
                argument,
            )
        logger.info(f'enqueue result for task {task_data.options.task_id}')
        asyncio.create_task(self.enqueue_result(task_data, future))

    def log_exception(self, task_data, e):
        """
        Logs an unhandled exception both with the runner logger
        and the extension one.
        """
        logger.warning(
            f'Got exception during execution of task {task_data.options.task_id}: {e}',
        )
        ext_logger = self.handler.get_logger(
            extra={'task_id': task_data.options.task_id},
        )
        ext_logger.exception(
            f'Unhandled exception during execution of task {task_data.options.task_id}',
        )

    @abstractmethod
    async def build_response(self, task_data, future):  # pragma: no cover
        """
        Wait for a task future to be completed, set the results
        on the task object and returns it.
        """
        pass

    @abstractmethod
    async def get_argument(self, task_data):  # pragma: no cover
        """
        Retrieves the Connect object source of the event that
        generates the task in order to pass it as the extension
        method argument.
        """
        pass

    async def get_installation(self, task_data):
        """
        Get the related event installation for multi-account
        extension.
        """
        return await self.client(
            'devops',
        ).services[self.config.service_id].installations[task_data.options.installation_id].get()

    @abstractmethod
    def get_method_name(self, task_data, argument):  # pragma: no cover
        """
        Returns the extension method has to be invoked.
        """
        pass



import asyncio
import inspect
import json
import logging
import os
import time
import traceback
from copy import (
    copy,
)
from tempfile import (
    NamedTemporaryFile,
)

import requests
from openpyxl import (
    Workbook,
    load_workbook,
)
from openpyxl.cell import (
    WriteOnlyCell,
)
from openpyxl.styles.named_styles import (
    NamedStyle,
)

from connect.client import (
    AsyncConnectClient,
    ConnectClient,
)
from connect.client.models import (
    AsyncResource,
)
from connect.eaas.core.enums import (
    ResultType,
)
from connect.eaas.core.logging import (
    RequestLogger,
)
from connect.eaas.core.proto import (
    Task,
    TaskOutput,
)
from connect.eaas.core.responses import (
    RowTransformationResponse,
    TransformationResponse,
)
from connect.eaas.runner.constants import (
    DOWNLOAD_CHUNK_SIZE,
    EXCEL_NULL_MARKER,
    ROW_DELETED_MARKER,
    TRANSFORMATION_TASK_MAX_PARALLEL_LINES,
    UPLOAD_CHUNK_SIZE,
)
from connect.eaas.runner.managers.base import (
    TasksManagerBase,
)
from connect.eaas.runner.managers.utils import (
    ResultStore,
)


logger = logging.getLogger(__name__)


class RowTransformationError(Exception):
    pass


class TransformationTasksManager(TasksManagerBase):
    """
    Class for Transformations managers.
    """

    def get_client(self, task_data):
        if task_data.options.api_key:
            return AsyncConnectClient(
                task_data.options.api_key,
                endpoint=self.config.get_api_url(),
                use_specs=False,
                default_headers=self.config.get_user_agent(),
                logger=RequestLogger(logger),
            )

        return self.client

    def get_extension_logger(self, task_data):
        return self.handler.get_logger(
            extra={'task_id': task_data.options.task_id},
        )

    def get_sync_client(self, task_data):
        return ConnectClient(
            task_data.options.api_key,
            endpoint=self.config.get_api_url(),
            use_specs=False,
            default_headers=self.config.get_user_agent(),
            logger=RequestLogger(logger),
        )

    def send_skip_response(self, data, output):
        future = asyncio.Future()
        future.set_result(TransformationResponse.skip(output))
        asyncio.create_task(self.enqueue_result(data, future))

    async def get_argument(self, task_data):
        """
        Get the transformation request through Connect public API
        related to the processing task.
        """
        client = self.get_client(task_data)
        object_exists = await self.filter_collection_by_event_definition(
            client,
            task_data,
        )
        if not object_exists:
            return

        definition = self.config.event_definitions[task_data.input.event_type]
        url = definition.api_resource_endpoint.format(pk=task_data.input.object_id)
        resource = AsyncResource(client, url)

        return await resource.get()

    def get_method_name(self, task_data, argument):
        return task_data.input.data['method']

    async def invoke(self, task_data, method, tfn_request):
        """
        Creates async task for transformation process.

        :param task_data: Data of the task to be processed.
        :type task_data: connect.eaas.core.proto.Task
        :param method: The method that has to be invoked.
        :param tfn_request: Connect tfn request that need to be processed.
        :type tfn_request: dict
        """

        future = asyncio.create_task(self.process_transformation(
            task_data, tfn_request, method,
        ))

        logger.info(f'Enqueue result for task {task_data.options.task_id}')
        asyncio.create_task(self.enqueue_result(task_data, future))

    async def build_response(self, task_data, future):
        result_message = Task(**task_data.dict())
        timeout = self.config.get_timeout('transformation')
        try:
            begin_ts = time.monotonic()
            result = await asyncio.wait_for(
                future,
                timeout=timeout,
            )
            result_message.output = TaskOutput(result=result.status)
            result_message.output.runtime = time.monotonic() - begin_ts
            logger.info(
                f'Transformation task {task_data.options.task_id} '
                f'result: {result.status}, '
                f'took: {result_message.output.runtime}',
            )
            if result.status in (ResultType.SKIP, ResultType.FAIL):
                result_message.output.message = result.output
        except Exception as e:
            cause = (
                str(e) if not isinstance(e, asyncio.TimeoutError)
                else 'timed out after {timeout} s'
            )
            self.log_exception(task_data, e)
            await self._fail_task(task_data, cause)
            result_message.output = TaskOutput(result=ResultType.FAIL)
            result_message.output.message = traceback.format_exc()[:4000]

        return result_message

    async def _fail_task(self, task_data, message):
        try:
            client = self.get_client(task_data)
            ns, _ = task_data.input.event_type.split('_', 1)
            await client(ns).requests[task_data.input.object_id]('fail').post()
            await client.conversations[task_data.input.object_id].messages.create(
                payload={
                    'type': 'message',
                    'text': f'Transformation request processing failed: {message}.',
                },
            )
        except Exception:
            logger.exception(f'Cannot fail the transformation request {task_data.input.object_id}')

    async def process_transformation(self, task_data, tfn_request, method):
        extension_logger = self.get_extension_logger(task_data)
        semaphore = asyncio.Semaphore(TRANSFORMATION_TASK_MAX_PARALLEL_LINES)
        input_file = await asyncio.get_running_loop().run_in_executor(
            self.executor,
            self.download_excel,
            tfn_request,
            task_data.options.api_key,
            extension_logger,
        )
        output_file = NamedTemporaryFile(
            suffix=f'.{tfn_request["files"]["input"]["name"].split(".")[-1]}',
        )

        read_queue = asyncio.Queue(TRANSFORMATION_TASK_MAX_PARALLEL_LINES)
        result_store = ResultStore()

        loop = asyncio.get_event_loop()

        reader_task = loop.run_in_executor(
            self.executor,
            self.read_excel,
            tfn_request,
            input_file,
            read_queue,
            extension_logger,
            loop,
        )
        writer_task = loop.run_in_executor(
            self.executor,
            self.write_excel,
            output_file.name,
            result_store,
            tfn_request['stats']['rows']['total'],
            tfn_request['transformation']['columns']['output'],
            task_data,
            extension_logger,
            loop,
        )
        processor_task = asyncio.create_task(self.process_rows(
            semaphore,
            read_queue,
            result_store,
            method,
            tfn_request,
            extension_logger,
        ))

        tasks = [reader_task, writer_task, processor_task]
        try:
            await asyncio.gather(*tasks)
        except Exception as e:
            logger.exception(
                f'Error during processing transformation '
                f'for {task_data.options.task_id}: {e}',
            )
            for task in tasks:
                if not task.done():
                    task.cancel()
            input_file.close()
            output_file.close()
            client = self.get_client(task_data)
            ns, _ = task_data.input.event_type.split('_', 1)
            await client(ns).requests[task_data.input.object_id]('fail').post()
            await client.conversations[task_data.input.object_id].messages.create(
                payload={
                    'type': 'message',
                    'text': f'Transformation request processing failed: {str(e)}',
                },
            )
            return TransformationResponse.fail(output=str(e))

        await self.send_output_file(
            task_data, tfn_request['batch']['id'], output_file, extension_logger,
        )
        input_file.close()
        output_file.close()
        return TransformationResponse.done()

    def download_excel(self, tfn_request, api_key, logger):
        input_file_name = tfn_request['files']['input']['name']
        input_file = NamedTemporaryFile(suffix=f'.{input_file_name.split(".")[-1]}')
        logger.info(
            f'Downloading input file for {tfn_request["id"]} '
            f'from {self.config.get_api_address()}{input_file_name}',
        )
        with requests.get(
            url=f'{self.config.get_api_address()}{input_file_name}',
            stream=True,
            headers={
                'Authorization': api_key,
                **self.config.get_user_agent(),
            },
        ) as response:
            response.raise_for_status()
            content_length = response.headers.get('Content-Length')
            progress = 0
            for chunk in response.iter_content(chunk_size=DOWNLOAD_CHUNK_SIZE):
                input_file.write(chunk)
                progress += len(chunk)
                logger.debug(
                    f'Input file download progress for {tfn_request["id"]}:'
                    f' {progress}/{content_length} bytes',
                )
        logger.info(
            f'Input file for {tfn_request["id"]} '
            f'from {self.config.get_api_address()}{input_file_name} downloaded',
        )
        return input_file

    def read_excel(self, tfn_request, filename, queue, logger, loop):
        wb = load_workbook(filename=filename, read_only=True)
        ws = wb['Data']
        lookup_columns = {}
        total_rows = tfn_request['stats']['rows']['total']
        delta = 1 if total_rows <= 10 else round(total_rows / 10)
        for idx, row in enumerate(ws.rows, start=1):
            if idx == 1:
                for col_idx, col_value in enumerate(row, start=1):
                    lookup_columns[col_idx] = col_value.value
                continue

            row_data = {col_name: None for col_name in lookup_columns.values()}
            row_styles = copy(row_data)

            for col_idx, col_value in enumerate(row, start=1):
                row_data[lookup_columns[col_idx]] = col_value.value
                style = NamedStyle()
                style.font = col_value.font
                style.fill = col_value.fill
                style.border = col_value.border
                style.number_format = col_value.number_format
                row_styles[lookup_columns[col_idx]] = style
            asyncio.run_coroutine_threadsafe(
                queue.put((idx, row_data, row_styles)),
                loop,
            )
            if idx % delta == 0 or idx == total_rows:
                logger.info(
                    f'Input file read progress for {tfn_request["id"]}:'
                    f' {idx}/{total_rows} rows',
                )

        logger.info(f'Input file read complete for {tfn_request["id"]}')
        wb.close()

    async def process_rows(self, semaphore, read_queue, result_store, method, tfn_request, logger):
        rows_processed = 0
        tasks = []
        total_rows = tfn_request['stats']['rows']['total']
        delta = 1 if total_rows <= 10 else round(total_rows / 10)
        while rows_processed < total_rows:
            await semaphore.acquire()
            row_idx, row, row_styles = await read_queue.get()
            if inspect.iscoroutinefunction(method):
                tasks.append(
                    asyncio.create_task(
                        asyncio.wait_for(
                            self.async_process_row(
                                semaphore,
                                method,
                                row_idx,
                                row,
                                row_styles,
                                result_store,
                            ),
                            self.config.get_timeout('row_transformation'),
                        ),
                    ),
                )
            else:
                loop = asyncio.get_running_loop()
                tasks.append(
                    asyncio.create_task(
                        asyncio.wait_for(
                            loop.run_in_executor(
                                self.executor,
                                self.sync_process_row,
                                semaphore,
                                method,
                                row_idx,
                                row,
                                row_styles,
                                result_store,
                                loop,
                            ),
                            self.config.get_timeout('row_transformation'),
                        ),
                    ),
                )

            rows_processed += 1
            if rows_processed % delta == 0 or rows_processed == total_rows:
                logger.info(
                    f'Starting transformation tasks for {tfn_request["id"]}:'
                    f' {rows_processed}/{total_rows} started',
                )

        try:
            logger.debug('gathering transformation tasks...')
            await asyncio.gather(*tasks)
        except Exception as e:
            logger.exception('Error during applying transformations.')
            for task in tasks:
                task.cancel()
            raise e

    async def async_process_row(self, semaphore, method, row_idx, row, row_styles, result_store):
        try:
            if ROW_DELETED_MARKER in list(row.values()):
                await result_store.put(row_idx, RowTransformationResponse.delete())
                return
            kwargs = {}
            if 'row_styles' in inspect.signature(method).parameters:
                kwargs['row_styles'] = row_styles
            response = await method(row, **kwargs)
            if not isinstance(response, RowTransformationResponse):
                raise RowTransformationError(f'invalid row tranformation response: {response}')
            if response.status == ResultType.FAIL:
                raise RowTransformationError(f'row transformation failed: {response.output}')
            await result_store.put(row_idx, response)
        except Exception as e:
            raise RowTransformationError(
                f'Error applying transformation function {method.__name__} '
                f'to row #{row_idx}: {str(e)}.',
            ) from e
        finally:
            semaphore.release()

    def sync_process_row(self, semaphore, method, row_idx, row, row_styles, result_store, loop):
        try:
            if ROW_DELETED_MARKER in list(row.values()):
                asyncio.run_coroutine_threadsafe(
                    result_store.put(row_idx, RowTransformationResponse.delete()), loop,
                )
                return
            kwargs = {}
            if 'row_styles' in inspect.signature(method).parameters:
                kwargs['row_styles'] = row_styles
            response = method(row, **kwargs)
            if not isinstance(response, RowTransformationResponse):
                raise RowTransformationError(f'invalid row tranformation response: {response}')
            if response.status == ResultType.FAIL:
                raise RowTransformationError(f'row transformation failed: {response.output}')
            asyncio.run_coroutine_threadsafe(result_store.put(row_idx, response), loop)
        except Exception as e:
            raise RowTransformationError(
                f'Error applying transformation function {method.__name__} '
                f'to row #{row_idx}: {str(e)}.',
            ) from e
        finally:
            semaphore.release()

    def write_excel(
        self, filename, result_store, total_rows, output_columns, task_data, logger, loop,
    ):
        wb = Workbook(write_only=True)

        ws = wb.create_sheet('Data')
        ws_columns = wb.create_sheet('Columns')
        ws_columns.append(['Name', 'Type', 'Nullable', 'Description', 'Precision'])
        column_keys = ['name', 'type', 'nullable', 'description', 'precision']

        column_names = []

        for column in output_columns:
            row = [column.get(key) for key in column_keys]
            ws_columns.append(row)
            column_names.append(column.get('name'))

        ws.append(column_names)

        rows_processed = 0
        delta = 1 if total_rows <= 10 else round(total_rows / 10)

        for idx in range(2, total_rows + 2):
            future = asyncio.run_coroutine_threadsafe(
                result_store.get(idx),
                loop,
            )
            response = future.result(
                timeout=self.config.env['transformation_write_queue_timeout'],
            )

            ws.append(self.generate_output_row(ws, column_names, response))
            rows_processed += 1
            if rows_processed % delta == 0 or rows_processed == total_rows:
                self.send_stat_update(task_data, rows_processed, total_rows)
                logger.info(
                    f'Writing to output file for {task_data.input.object_id}: {rows_processed}/'
                    f'{total_rows} written',
                )

        wb.save(filename)

    def generate_output_row(self, ws, column_names, response):
        row = []
        for col_name in column_names:
            if response.status == ResultType.SUCCESS:
                value = response.transformed_row.get(col_name)
                cell = WriteOnlyCell(ws=ws, value=value if value is not None else EXCEL_NULL_MARKER)
                style = response.transformed_row_styles.get(col_name, NamedStyle())
                cell.number_format = style.number_format
                cell.font = style.font
                cell.fill = style.fill
                cell.border = style.border
                row.append(cell)
            elif response.status == ResultType.SKIP:
                row.append(EXCEL_NULL_MARKER)
            elif response.status == ResultType.DELETE:
                row.append(ROW_DELETED_MARKER)
            else:
                raise RowTransformationError(
                    f'Invalid row transformation response status: {response.status}.',
                )
        return row

    def send_stat_update(self, task_data, rows_processed, total_rows):
        client = self.get_sync_client(task_data)
        ns, _ = task_data.input.event_type.split('_', 1)
        client(ns).requests[task_data.input.object_id].update(
            payload={'stats': {'rows': {'total': total_rows, 'processed': rows_processed}}},
        )

    async def send_output_file(self, task_data, batch_id, output_file, logger):
        client = self.get_client(task_data)

        fileobj = open(output_file.name, 'rb')
        fileobj.seek(0, os.SEEK_END)
        file_size = fileobj.tell()
        fileobj.seek(0)

        filename = f'{task_data.input.object_id}-out.{output_file.name.split(".")[-1]}'
        headers = {
            'Content-Type': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
            'Content-Disposition': f'attachment; filename="{filename}"',
            'Content-Length': str(file_size),
        }

        async def chunks_iterator():    # pragma: no cover
            progress = 0
            while data := fileobj.read(UPLOAD_CHUNK_SIZE):
                yield data
                progress += len(data)
                logger.debug(
                    f'Output file upload progress for {task_data.input.object_id}:'
                    f' {progress}/{file_size} bytes',
                )

        media_file = await client.ns('media').ns('folders').collection(
            'streams_batches',
        )[batch_id].action(
            'files',
        ).post(
            content=chunks_iterator(),
            headers=headers,
        )
        logger.info(
            f'Output file upload completed for {task_data.input.object_id}',
        )
        media_file_id = json.loads(media_file)['id']
        ns, _ = task_data.input.event_type.split('_', 1)
        await client(ns).requests[task_data.input.object_id].update(
            payload={'files': {'output': {'id': media_file_id}}},
        )
        await client(ns).requests[task_data.input.object_id]('process').post()
